# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
#
# Copyright (c) 2024 Stefan Bender
#
# This module is part of pyspaceweather.
# pyspaceweather is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying COPYING.GPLv2 file or http://www.gnu.org/licenses/gpl-2.0.html.
"""Space weather index read tests for GFZ formats

Parsing tests for the GFZ file formats.
"""
import os

import numpy as np
import pandas as pd
import requests

import pytest

from spaceweather import (
	gfz_3h, gfz_daily, get_gfz_age, update_gfz,
)
from spaceweather.gfz import GFZ_URL_30D, HP30_URL_30D, HP60_URL_30D

GFZ_PATH_ALL = os.path.join("tests", "Kp_ap_Ap_SN_F107_since_2024.txt")
GFZ_PATH_30D = os.path.join("tests", "Kp_ap_Ap_SN_F107_nowcast.txt")

HP30_PATH_ALL = os.path.join("tests", "Hp30_ap30_complete_series.txt")
HP30_PATH_30D = os.path.join("tests", "Hp30_ap30_nowcast.txt")

HP60_PATH_ALL = os.path.join("tests", "Hp60_ap60_complete_series.txt")
HP60_PATH_30D = os.path.join("tests", "Hp60_ap60_nowcast.txt")


@pytest.fixture(scope="module")
def df_3h():
	return gfz_3h(gfzpath_all=GFZ_PATH_ALL, gfzpath_30d=GFZ_PATH_30D)


def test_age():
	now = pd.Timestamp.utcnow()
	for p in [GFZ_PATH_ALL, GFZ_PATH_30D]:
		assert os.path.exists(p)
		fage0 = get_gfz_age(p)
		fage1 = now - get_gfz_age(p, relative=False)
		assert (fage0 > pd.Timedelta("3h")) == (fage1 > pd.Timedelta("3h"))
		assert (fage0 > pd.Timedelta("1d")) == (fage1 > pd.Timedelta("1d"))


def test_update(mocker):
	mocker.patch("requests.get")
	update_gfz(min_age="1d", gfzpath_all=GFZ_PATH_ALL, gfzpath_30d=GFZ_PATH_30D)
	requests.get.assert_called_with(GFZ_URL_30D, stream=True)


def test_auto_update(mocker, tmpdir):
	# test with non-existent file
	mocker.patch("requests.get")
	tmpdir = str(tmpdir)
	update_gfz(gfzpath_30d=os.path.join(tmpdir, "foo.dat"))
	requests.get.assert_called_with(GFZ_URL_30D, stream=True)
	# Should update the last-5-year data
	gfz_daily(
		gfzpath_all=GFZ_PATH_ALL, gfzpath_30d=GFZ_PATH_30D,
		update=True, update_interval="1d",
	)
	requests.get.assert_called_with(GFZ_URL_30D, stream=True)
	with pytest.warns(UserWarning):
		gfz_daily(
			gfzpath_all=GFZ_PATH_ALL, gfzpath_30d=GFZ_PATH_30D,
			update=False, update_interval="0h",
		)


def test_not_avail(mocker, tmpdir):
	# test with non-existent file
	mocker.patch("requests.get")
	tmpdir = str(tmpdir)
	tmpfile = os.path.join(tmpdir, "foo.dat")
	# daily
	with pytest.raises(IOError):
		with pytest.warns(UserWarning):
			gfz_daily(update=False, update_interval="0h", gfzpath_30d=tmpfile)
	# 3h data
	with pytest.raises(IOError):
		with pytest.warns(UserWarning):
			gfz_3h(update=False, update_interval="0h", gfzpath_30d=tmpfile)


def test_daily():
	df = gfz_daily(gfzpath_all=GFZ_PATH_ALL, gfzpath_30d=GFZ_PATH_30D)
	np.testing.assert_allclose(
		df.loc["2024-01-01"].values,
		np.array([
			2.024e+03, 1.000e+00, 1.000e+00,
			3.36030e+04, 3.36035e+04, 2.596e+03, 2.500e+01,
			6.670e-01, 3.330e-01, 6.670e-01, 1.333e+00,
			2.000e+00, 3.000e+00, 3.333e+00, 4.000e+00, 1.5333e+01,
			3.000e+00, 2.000e+00, 3.000e+00, 5.000e+00,
			7.000e+00, 1.500e+01, 1.800e+01, 2.700e+01, 1.000e+01,
			5.400e+01, 1.357e+02, 1.312e+02, 1.000e+00,
		], dtype=np.float64),
		rtol=1e-6,
	)


@pytest.mark.parametrize(
	"name, result",
	[
		("Ap", [3, 2, 3, 5, 7, 15, 18, 27]),
		("Kp", [0.667, 0.333, 0.667, 1.333, 2.0, 3.0, 3.333, 4.0]),
	]
)
def test_3hourly_index(name, result, df_3h):
	df = df_3h
	np.testing.assert_allclose(
		df.loc[
			pd.date_range(
				"2024-01-01 01:30", "2024-01-01 23:30", freq="3h"
			)
		][name].values,
		np.array(result, dtype=np.float64),
		rtol=1e-6,
	)


@pytest.mark.parametrize(
	"fpall, fp30d, url",
	[
		(HP30_PATH_ALL, HP30_PATH_30D, HP30_URL_30D),
		(HP60_PATH_ALL, HP60_PATH_30D, HP60_URL_30D),
	],
	ids=["Hp30", "Hp60"],
)
def test_auto_update_hp(fpall, fp30d, url, mocker, request):
	mocker.patch("requests.get")
	_gfz_fmt = request.node.callspec.id.lower()
	# Should update the last-5-year data
	gfz_daily(
		gfzpath_all=fpall, gfzpath_30d=fp30d,
		update=True, update_interval="1d",
		gfz_format=_gfz_fmt,
	)
	requests.get.assert_called_with(url, stream=True)


@pytest.mark.parametrize(
	"fpall, fp30d, index, expected",
	[
		(
			HP30_PATH_ALL, HP30_PATH_30D, "2025-07-01 00:15",
			[2025, 7, 1, 0, 0.25, 34150.0, 34150.01042, 3.000, 15, 0],
		),
		(
			HP60_PATH_ALL, HP60_PATH_30D, "2025-07-01 00:30",
			[2025, 7, 1, 0, 0.50, 34150.0, 34150.02083, 3.333, 18, 0],
		),
	],
	ids=["Hp30", "Hp60"],
)
def test_daily_hp(fpall, fp30d, index, expected, request):
	_gfz_fmt = request.node.callspec.id.lower()
	df = gfz_daily(
		gfzpath_all=fpall,
		gfzpath_30d=fp30d,
		gfz_format=_gfz_fmt,
	)
	np.testing.assert_allclose(
		df.loc[index].values,
		np.array(expected, dtype=np.float64),
		rtol=1e-6,
	)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from os import getenv
from glob import glob

from spaceweather import __version__

project = u'pyspaceweather'
copyright = u'2020, Stefan Bender'
author = u'Stefan Bender'

version = __version__
release = __version__

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.extlinks',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx.ext.napoleon',
    'sphinx.ext.mathjax',
    'recommonmark',
    'numpydoc'
]
if getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_GB'

master_doc = 'index'

language = "en"

htmlhelp_basename = 'pyspaceweatherdoc'

# autodoc_docstring_signature = False
# autodoc_dumb_docstring = True
autosummary_generate = True

latex_elements = {
    "papersize": "a4paper",
    "pointsize": "11pt",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'pyspaceweather.tex', u'pyspaceweather Documentation',
     u'Stefan Bender', 'manual'),
]

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'pyspaceweather', u'pyspaceweather Documentation',
     [author], 1)
]

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'pyspaceweather', u'pyspaceweather Documentation',
     author, 'pyspaceweather', 'Space weather indices for python.',
     'Miscellaneous'),
]

napoleon_use_admonition_for_notes = True

html_theme = "sphinx_rtd_theme"

# on_rtd is whether we are on readthedocs.org
on_rtd = getenv("READTHEDOCS", None) == "True"
if not on_rtd:
    import sphinx_rtd_theme
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

exclude_patterns = [u'_build', 'Thumbs.db', '.DS_Store']
pygments_style = 'default'
templates_path = ['_templates']
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_split_index = True
html_sidebars = {
    '**': ['searchbox.html', 'localtoc.html', 'relations.html',
          'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)
html_context = dict(
    display_github=True,
    github_user="st-bender",
    github_repo="pyspaceweather",
    github_version="master",
    conf_py_path="/docs/",
)
html_extra_path = glob("../COPYING*") + glob("../*.txt")
# Switch to old behavior with html4, for a good display of references,
# as described in https://github.com/sphinx-doc/sphinx/issues/6705
# html4_writer = True

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True
# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'pandas': ('https://pandas.pydata.org/pandas-docs/stable/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
}



# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
#
# Copyright (c) 2022 Stefan Bender
#
# This module is part of pyspaceweather.
# pyspaceweather is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying COPYING.GPLv2 file or http://www.gnu.org/licenses/gpl-2.0.html.
"""OMNI data read tests
"""
import os
import requests
from posixpath import join as urljoin

import numpy as np
import pandas as pd

import pytest

from spaceweather import cache_omnie, omnie_hourly, omnie_mask_missing, sw_daily
from spaceweather.omni import OMNI_URL_BASE, OMNI_PREFIX, OMNI_EXT

_TEST_YEAR = 2012
_TEST_FILE = "{0}_{1:04d}.{2}".format(OMNI_PREFIX, _TEST_YEAR, OMNI_EXT)
_TEST_URL = urljoin(OMNI_URL_BASE, _TEST_FILE)
_TEST_PATH = os.path.join(".", "tests")


@pytest.fixture(scope="module")
def df_d():
	return sw_daily()


@pytest.fixture(scope="module")
def df_o():
	return omnie_hourly(2000, local_path=_TEST_PATH, prefix="omni2t")


def test_cache(mocker, tmpdir):
	mocker.patch("requests.get")
	tmpdir = str(tmpdir)
	# Check non-existent file
	cache_omnie(year=_TEST_YEAR, local_path=tmpdir)
	requests.get.assert_called_once_with(_TEST_URL, stream=True)
	# Check non-existent (sub)dir
	tmppath = os.path.join(tmpdir, "data")
	cache_omnie(year=_TEST_YEAR, local_path=tmppath)
	requests.get.assert_called_with(_TEST_URL, stream=True)


def test_auto_update(mocker, tmpdir):
	# test with non-existent file
	mocker.patch("requests.get")
	tmpdir = str(tmpdir)
	with pytest.raises(IOError):
		with pytest.warns(UserWarning):
			omnie_hourly(year=_TEST_YEAR, cache=True, local_path=tmpdir)
	requests.get.assert_called_once_with(_TEST_URL, stream=True)


def test_not_avail(mocker, tmpdir):
	# test with non-existent file
	tmpdir = str(tmpdir)
	with pytest.raises(IOError):
		with pytest.warns(UserWarning):
			omnie_hourly(year=_TEST_YEAR, cache=False, local_path=tmpdir)


@pytest.mark.parametrize("hour", range(0, 24, 3))
@pytest.mark.parametrize("index", ["Ap", "Kp"])
def test_hourly(hour, index, df_d, df_o):
	df1 = df_o
	# The last row is for the missing value test.
	df1 = df1.iloc[:-1]
	df2 = df_d
	ind_name = "{0}{1}".format(index, hour)
	df1_ind = df1[df1["hour"] == hour][index]
	df2_ind = df2[ind_name].loc[df1_ind.index.date].rename(index)
	df2_ind.index = df1_ind.index
	pd.testing.assert_series_equal(df1_ind, df2_ind)


@pytest.mark.parametrize(
	"name, result",
	[
		("Ap", np.array([56, 39, 27, 18, 32, 15, 32, 22])),
		("Kp", np.array([5.3, 4.7, 4.0, 3.3, 4.3, 3.0, 4.3, 3.7])),
	]
)
def test_3hourly_index(name, result, df_o):
	df = df_o
	np.testing.assert_allclose(
		df.loc[
			pd.date_range(
				"2000-01-01 00:00", "2000-01-01 23:00", freq="3h"
			)
		][name].values,
		result,
		rtol=1e-12,
	)


def test_mask_missing(df_o):
	df = df_o
	dfp = omnie_mask_missing(df)
	# The last row should contain all NaNs.
	dfp = dfp.iloc[-1]
	for v in filter(
		lambda n: n not in ["year", "doy", "hour"],
		dfp.index,
	):
		assert np.isnan(dfp[v])



import pytest


def pytest_addoption(parser):
	parser.addoption(
		"--run-online",
		action="store_true",
		default=False,
		help="Run online tests",
	)


def pytest_collection_modifyitems(config, items):
	if not config.getoption("--run-online"):
		skipper = pytest.mark.skip(reason="Only run when --run-online is given")
		for item in items:
			if "online" in item.keywords:
				item.add_marker(skipper)



# Copyright (c) 2020--2022 Stefan Bender
#
# This module is part of pyspaceweather.
# pyspaceweather is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying COPYING.GPLv2 file or http://www.gnu.org/licenses/gpl-2.0.html.
"""Python interface for space weather indices

"""
__version__ = "0.4.0"

from .celestrak import *
from .gfz import *
from .omni import *



# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8
#
# Copyright (c) 2020 Stefan Bender
#
# This module is part of pyspaceweather.
# pyspaceweather is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying COPYING.GPLv2 file or http://www.gnu.org/licenses/gpl-2.0.html.
"""Space weather index read tests
"""
import os
import shutil

import numpy as np
import pandas as pd
import requests

import pytest

from spaceweather import (
	ap_kp_3h, sw_daily, get_file_age, update_data,
	SW_PATH_ALL, SW_PATH_5Y,
)
from spaceweather.celestrak import DL_URL_5Y


@pytest.fixture(scope="module")
def df_3h():
	return ap_kp_3h()


def test_age():
	now = pd.Timestamp.utcnow()
	for p in [SW_PATH_ALL, SW_PATH_5Y]:
		assert os.path.exists(p)
		fage0 = get_file_age(p)
		fage1 = now - get_file_age(p, relative=False)
		assert (fage0 > pd.Timedelta("3h")) == (fage1 > pd.Timedelta("3h"))
		assert (fage0 > pd.Timedelta("1d")) == (fage1 > pd.Timedelta("1d"))


def _assert_age(p, age):
	assert os.path.exists(p)
	fage = get_file_age(p)
	assert fage < pd.Timedelta(age)


@pytest.mark.online
def test_update(tmpdir):
	# test with temporary files
	tmpdir = str(tmpdir)
	tmp_all = os.path.join(tmpdir, "tmp_all.dat")
	tmp_5y = os.path.join(tmpdir, "tmp_5y.dat")
	shutil.copy(SW_PATH_ALL, tmp_all)
	shutil.copy(SW_PATH_5Y, tmp_5y)
	update_data(swpath_all=tmp_all, swpath_5y=tmp_5y, min_age="100d")
	for (p, age) in zip([tmp_all, tmp_5y], ["1460d", "100d"]):
		_assert_age(p, age)


def test_auto_update(mocker, tmpdir):
	# test with non-existent file
	mocker.patch("requests.get")
	tmpdir = str(tmpdir)
	update_data(swpath_5y=os.path.join(tmpdir, "foo.dat"))
	requests.get.assert_called_with(DL_URL_5Y, stream=True)
	# Should update the last-5-year data
	sw_daily(update=True, update_interval="1d")
	requests.get.assert_called_with(DL_URL_5Y, stream=True)
	with pytest.warns(UserWarning):
		sw_daily(update=False, update_interval="0h")


def test_not_avail(mocker, tmpdir):
	# test with non-existent file
	mocker.patch("requests.get")
	tmpdir = str(tmpdir)
	tmpfile = os.path.join(tmpdir, "foo.dat")
	# daily
	with pytest.raises(IOError):
		with pytest.warns(UserWarning):
			sw_daily(update=False, update_interval="0h", swpath_5y=tmpfile)
	# 3h data
	with pytest.raises(IOError):
		with pytest.warns(UserWarning):
			ap_kp_3h(update=False, update_interval="0h", swpath_5y=tmpfile)


def test_daily():
	df = sw_daily()
	np.testing.assert_allclose(
		df.loc["2000-01-01"].values,
		np.array([
			2.000e+03, 1.000e+00, 1.000e+00, 2.272e+03, 7.000e+00, 5.300e+00, 4.700e+00,
			4.000e+00, 3.300e+00, 4.300e+00, 3.000e+00, 4.300e+00, 3.700e+00, 3.270e+01,
			5.600e+01, 3.900e+01, 2.700e+01, 1.800e+01, 3.200e+01, 1.500e+01, 3.200e+01,
			2.200e+01, 3.000e+01, 1.300e+00, 6.000e+00, 7.100e+01, 1.256e+02, 0.000e+00,
			1.611e+02, 1.750e+02, 1.299e+02, 1.662e+02, 1.790e+02,
		]),
		rtol=1e-12,
	)


@pytest.mark.parametrize(
	"name, result",
	[
		("Ap", np.array([56, 39, 27, 18, 32, 15, 32, 22])),
		("Kp", np.array([5.3, 4.7, 4.0, 3.3, 4.3, 3.0, 4.3, 3.7])),
	]
)
def test_3hourly_index(name, result, df_3h):
	df = df_3h
	np.testing.assert_allclose(
		df.loc[
			pd.date_range(
				"2000-01-01 01:30", "2000-01-01 23:30", freq="3h"
			)
		][name].values,
		result,
		rtol=1e-12,
	)



# Copyright (c) 2020--2022 Stefan Bender
#
# This module is part of pyspaceweather.
# pyspaceweather is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying COPYING.GPLv2 file or http://www.gnu.org/licenses/gpl-2.0.html.
"""Python interface for space weather indices

General file handling functions for space weather data
"""
import errno
import os
import warnings

import requests


def _assert_file_exists(f):
	if not os.path.exists(f):
		raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), f)


def _dl_file(swpath, url):
	with requests.get(url, stream=True) as r:
		if r.status_code != requests.codes.ok:
			if isinstance(r.status_code, int):
				warnings.warn(
					"Failed to download from {0}, status code: {1}".format(
						url, r.status_code,
					),
				)
			return
		with open(swpath, 'wb') as fd:
			for chunk in r.iter_content(chunk_size=1024):
				fd.write(chunk)


def _resource_filepath(file, subdir="data"):
	try:
		from contextlib import ExitStack
		from importlib import resources
		file_manager = ExitStack()
		ref = resources.files(__package__) / subdir / file
		filepath = file_manager.enter_context(resources.as_file(ref))
	except (AttributeError, ImportError):
		from pkg_resources import resource_filename
		filepath = resource_filename(__package__, os.path.join(subdir, file))
	return filepath



# Copyright (c) 2020--2022 Stefan Bender
#
# This module is part of pyspaceweather.
# pyspaceweather is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying COPYING.GPLv2 file or http://www.gnu.org/licenses/gpl-2.0.html.
"""Python interface for space weather indices

Celestrak space weather indices file parser for python [#]_.

.. [#] https://celestrak.com/SpaceData/
"""
import os
import logging
from warnings import warn

import numpy as np
import pandas as pd

from .core import _assert_file_exists, _dl_file, _resource_filepath

__all__ = [
	"sw_daily", "ap_kp_3h", "read_sw",
	"get_file_age", "update_data",
	"SW_PATH_ALL", "SW_PATH_5Y",
]

DL_URL_ALL = "https://celestrak.com/SpaceData/SW-All.txt"
DL_URL_5Y = "https://celestrak.com/SpaceData/SW-Last5Years.txt"
SW_FILE_ALL = os.path.basename(DL_URL_ALL)
SW_FILE_5Y = os.path.basename(DL_URL_5Y)
SW_PATH_ALL = _resource_filepath(SW_FILE_ALL)
SW_PATH_5Y = _resource_filepath(SW_FILE_5Y)


def get_file_age(swpath, relative=True):
	"""Age of the downloaded data file

	Retrieves the last update time of the given file or full path.

	Parameters
	----------
	swpath: str
		Filename to check, absolute path or relative to the current dir.
	relative: bool, optional, default True
		Return the file's age (True) or the last update time (False).

	Returns
	-------
	upd: pandas.Timestamp or pandas.Timedelta
		The last updated time or the file age, depending on the setting
		of `relative` above.
		Raises ``IOError`` if the file is not found.
	"""
	_assert_file_exists(swpath)
	with open(swpath) as fp:
		for line in fp:
			if line.startswith("UPDATED"):
				# closes the file automatically
				break
	upd = pd.to_datetime(line.lstrip("UPDATED"), utc=True)
	if relative:
		return pd.Timestamp.utcnow() - upd
	return upd


def update_data(
	min_age="3h",
	swpath_all=None, swpath_5y=None,
	url_all=None, url_5y=None,
):
	"""Update the local space weather index data

	Updates the local space weather index data from the website [#]_,
	given that the 5-year file is older
	than `min_age`, or the combined (large) file is older than four years.
	If the data is missing for some reason, a download will be attempted nonetheless.

	All arguments are optional and changing them from the defaults should not
	be required neither should it be necessary nor is it recommended.

	.. [#] https://celestrak.com/SpaceData/

	Parameters
	----------
	min_age: str, optional, default "3h"
		The time after which a new download will be attempted.
		The online data is updated every 3 hours, thus setting this value to
		a shorter time is not needed and not recommended.
	swpath_all: `None` or str, optional, default `None`
		Filename for the large combined index file including the
		historic data, absolute path or relative to the current dir.
		`None` uses the package's default file location.
	swpath_5y: `None` or str, optional, default `None`
		Filename for the 5-year index file, absolute path or relative to the current dir.
		`None` uses the package's default file location.
	url_all: `None` or str, optional, default `None`
		The url of the "historic" data file.
		`None` uses the default url.
	url_5y: `None` or str, optional, default `None`
		The url of the data file of containing the indices of the last 5 years.
		`None` uses the default url.

	Returns
	-------
	Nothing.
	"""
	def _update_file(swpath, url, min_age):
		if not os.path.exists(swpath):
			logging.info("{0} not found, downloading.".format(swpath))
			_dl_file(swpath, url)
			return
		if get_file_age(swpath) < pd.Timedelta(min_age):
			logging.info("not updating '{0}'.".format(swpath))
			return
		logging.info("updating '{0}'.".format(swpath))
		_dl_file(swpath, url)

	swpath_all = swpath_all or SW_PATH_ALL
	swpath_5y = swpath_5y or SW_PATH_5Y
	url_all = url_all or DL_URL_ALL
	url_5y = url_5y or DL_URL_5Y

	# Update the large file after four years
	# to have some overlap with the 5-year data
	# 1460 = 4 * 365
	_update_file(swpath_all, url_all, "1460days")
	# Don't re-download before `min_age` has passed (3h)
	_update_file(swpath_5y, url_5y, min_age)


def read_sw(swpath):
	"""Read and parse space weather index data file

	Reads the given file and parses it according to the space weather data format.

	Parameters
	----------
	swpath: str
		File to parse, absolute path or relative to the current dir.

	Returns
	-------
	sw_df: pandas.DataFrame
		The parsed space weather data (daily values).
		Raises an ``IOError`` if the file is not found.
		The index is returned timezone-naive but contains UTC timestamps.
		To convert to a timezone-aware index, use
		:meth:`pandas.DataFrame.tz_localize()`: ``sw_df.tz_localize("utc")``.

		The dataframe contains the following columns:

		"year", "month", "day":
			The observation date
		"bsrn":
			Bartels Solar Rotation Number.
			A sequence of 27-day intervals counted continuously from 1832 Feb 8.
		"rotd":
			Number of Day within the Bartels 27-day cycle (01-27).
		"Kp0", "Kp3", "Kp6", "Kp9", "Kp12", "Kp15", "Kp18", "Kp21":
			Planetary 3-hour Range Index (Kp) for 0000-0300, 0300-0600,
			0600-0900, 0900-1200, 1200-1500, 1500-1800, 1800-2100, 2100-2400 UT
		"Kpsum": Sum of the 8 Kp indices for the day.
			Expressed to the nearest third of a unit.
		"Ap0", "Ap3", "Ap6", "Ap9", "Ap12", "Ap15", "Ap18", "Ap21":
			Planetary Equivalent Amplitude (Ap) for 0000-0300, 0300-0600,
			0600-0900, 0900-1200, 1200-1500, 1500-1800, 1800-2100, 2100-2400 UT
		"Apavg":
			Arithmetic average of the 8 Ap indices for the day.
		"Cp":
			Cp or Planetary Daily Character Figure. A qualitative estimate of
			overall level of magnetic activity for the day determined from the sum
			of the 8 Ap indices. Cp ranges, in steps of one-tenth, from 0 (quiet)
			to 2.5 (highly disturbed). "C9":
		"isn":
			International Sunspot Number.
			Records contain the Zurich number through 1980 Dec 31 and the
			International Brussels number thereafter.
		"f107_adj":
			10.7-cm Solar Radio Flux (F10.7) Adjusted to 1 AU.
			Measured at Ottawa at 1700 UT daily from 1947 Feb 14 until
			1991 May 31 and measured at Penticton at 2000 UT from 1991 Jun 01 on.
			Expressed in units of 10-22 W/m2/Hz.
		"Q":
			Flux Qualifier.
			0 indicates flux required no adjustment;
			1 indicates flux required adjustment for burst in progress at time of measurement;
			2 indicates a flux approximated by either interpolation or extrapolation;
			3 indicates no observation; and
			4 indicates CSSI interpolation of missing data.
		"f107_81ctr_adj":
			Centered 81-day arithmetic average of F10.7 (adjusted).
		"f107_81lst_adj":
			Last 81-day arithmetic average of F10.7 (adjusted).
		"f107_obs":
			Observed (unadjusted) value of F10.7.
		"f107_81ctr_obs":
			Centered 81-day arithmetic average of F10.7 (observed).
		"f107_81lst_obs":
			Last 81-day arithmetic average of F10.7 (observed).
	"""
	_assert_file_exists(swpath)
	sw = np.genfromtxt(
		swpath,
		skip_header=3,
		delimiter=[
		#  yy mm dd br rd kp kp kp kp kp kp kp kp Kp
			4, 3, 3, 5, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4,
		#  ap ap ap ap ap ap ap ap Ap cp c9 is f1  q
			4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 2, 4, 6, 2,
		#  f2 f3 f4 f5 f6
			6, 6, 6, 6, 6],
		dtype=(
			"i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,"
			"i4,i4,i4,i4,i4,i4,i4,i4,i4,f8,i4,i4,f8,i4,"
			"f8,f8,f8,f8,f8"
		),
		names=[
			"year", "month", "day", "bsrn", "rotd",
			"Kp0", "Kp3", "Kp6", "Kp9", "Kp12", "Kp15", "Kp18", "Kp21", "Kpsum",
			"Ap0", "Ap3", "Ap6", "Ap9", "Ap12", "Ap15", "Ap18", "Ap21", "Apavg",
			"Cp", "C9", "isn", "f107_adj", "Q", "f107_81ctr_adj", "f107_81lst_adj",
			"f107_obs", "f107_81ctr_obs", "f107_81lst_obs"
		]
	)[2:-1]
	sw = sw[sw["year"] != -1]
	ts = pd.to_datetime([
		"{0:04d}-{1:02d}-{2:02d}".format(yy, mm, dd)
		for yy, mm, dd in sw[["year", "month", "day"]]
	])
	sw_df = pd.DataFrame(sw, index=ts)
	# Adjust Kp to 0...9
	kpns = list(map("Kp{0}".format, range(0, 23, 3))) + ["Kpsum"]
	sw_df[kpns] = 0.1 * sw_df[kpns]
	return sw_df


# Common arguments for the public daily and 3h interfaces
_SW_COMMON_PARAMS = """
Parameters
----------
swpath_all: `None` or str, optional, default `None`
	Filename for the large combined index file including the
	historic data, absolute path or relative to the current dir.
	`None` uses the package's default file location.
swpath_5y: `None` or str, optional, default `None`
	Filename for the 5-year index file, absolute path or relative to the current dir.
	`None` uses the package's default file location.
update: bool, optional, default False
	Attempt to update the local data if it is older than `update_interval`.
update_interval: str, optional, default "30days"
	The time after which the data are considered "old".
	By default, no automatic re-download is initiated, set `update` to true.
	The online data is updated every 3 hours, thus setting this value to
	a shorter time is not needed and not recommended.
"""


def _doc_param(**sub):
	def dec(obj):
		obj.__doc__ = obj.__doc__.format(**sub)
		return obj
	return dec


@_doc_param(params=_SW_COMMON_PARAMS)
def sw_daily(swpath_all=None, swpath_5y=None, update=False, update_interval="30days"):
	"""Combined daily Ap, Kp, and f10.7 index values

	Combines the "historic" and last-5-year data into one dataframe.

	All arguments are optional and changing them from the defaults should not
	be required neither should it be necessary nor is it recommended.
	{params}
	Returns
	-------
	sw_df: pandas.DataFrame
		The combined parsed space weather data (daily values).
		Raises ``IOError`` if the data files cannot be found.
		The index is returned timezone-naive but contains UTC timestamps.
		To convert to a timezone-aware index, use
		:meth:`pandas.DataFrame.tz_localize()`: ``sw_df.tz_localize("utc")``.

	See Also
	--------
	ap_kp_3h, read_sw
	"""
	swpath_all = swpath_all or SW_PATH_ALL
	swpath_5y = swpath_5y or SW_PATH_5Y

	# ensure that the file exists and is up to date
	if (
		not os.path.exists(swpath_all)
		or not os.path.exists(swpath_5y)
	):
		warn("Could not find space weather data, trying to download.")
		update_data(swpath_all=swpath_all, swpath_5y=swpath_5y)

	if (
		# 1460 = 4 * 365
		get_file_age(swpath_all) > pd.Timedelta("1460days")
		or get_file_age(swpath_5y) > pd.Timedelta(update_interval)
	):
		if update:
			update_data(swpath_all=swpath_all, swpath_5y=swpath_5y)
		else:
			warn(
				"Local data files are older than {0}, pass `update=True` or "
				"run `sw.update_data()` manually if you need newer data.".format(
					update_interval
				)
			)

	df_all = read_sw(swpath_all)
	df_5y = read_sw(swpath_5y)
	return pd.concat([df_all[:df_5y.index[0]], df_5y[1:]])


@_doc_param(params=_SW_COMMON_PARAMS)
def ap_kp_3h(*args, **kwargs):
	"""3h values of Ap and Kp

	Provides the 3-hourly Ap and Kp indices from the full daily data set.

	Accepts the same arguments as `sw_daily()`.
	All arguments are optional and changing them from the defaults should not
	be required neither should it be necessary nor is it recommended.
	{params}
	Returns
	-------
	sw_df: pandas.DataFrame
		The combined Ap and Kp index data (3h values).
		The index values are centred at the 3h interval, i.e. at 01:30:00,
		04:30:00, 07:30:00, ... and so on.
		Raises ``IOError`` if the data files cannot be found.
		The index is returned timezone-naive but contains UTC timestamps.
		To convert to a timezone-aware index, use
		:meth:`pandas.DataFrame.tz_localize()`: ``sw_df.tz_localize("utc")``.

	See Also
	--------
	sw_daily
	"""
	daily_df = sw_daily(*args, **kwargs)
	ret = daily_df.copy()
	apns = list(map("Ap{0}".format, range(0, 23, 3)))
	kpns = list(map("Kp{0}".format, range(0, 23, 3)))
	for i, (ap, kp) in enumerate(zip(apns, kpns)):
		ret[ap].index = daily_df[ap].index + pd.Timedelta((i * 3 + 1.5), unit="h")
		ret[kp].index = daily_df[kp].index + pd.Timedelta((i * 3 + 1.5), unit="h")
	sw_ap = pd.concat(map(ret.__getitem__, apns))
	sw_kp = pd.concat(map(ret.__getitem__, kpns))
	df = pd.DataFrame({"Ap": sw_ap, "Kp": sw_kp})
	return df.reindex(df.index.sort_values())



# -*- coding: utf-8 -*-
# Copyright (c) 2022 Stefan Bender
#
# This module is part of pyspaceweather.
# pyspaceweather is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying COPYING.GPLv2 file or http://www.gnu.org/licenses/gpl-2.0.html.
"""Python interface for OMNI space weather data

Omni2 [#]_ space weather data interface for python.

.. [#] https://omniweb.gsfc.nasa.gov/ow.html
"""
import os
import logging
from warnings import warn

from posixpath import join as urljoin

import numpy as np
import pandas as pd

from .core import _assert_file_exists, _dl_file, _resource_filepath

__all__ = [
	"cache_omnie",
	"omnie_hourly",
	"omnie_mask_missing",
	"read_omnie",
]

OMNI_URL_BASE = "https://spdf.gsfc.nasa.gov/pub/data/omni/low_res_omni/extended"
OMNI_PREFIX, OMNI_EXT = "omni2", "dat"
OMNI_SUBDIR = "omni_extended"
LOCAL_PATH = _resource_filepath(OMNI_SUBDIR)

_OMNI_MISSING = {
	"year": None,
	"doy": None,
	"hour": None,
	"bsrn": 9999,
	"id_imf": 99,
	"id_sw": 99,
	"n_imf": 999,
	"n_plasma": 999,
	"B_mag_avg": 999.9,
	"B_mag": 999.9,
	"theta_B": 999.9,
	"phi_B": 999.9,
	"B_x": 999.9,
	"B_y_GSE": 999.9,
	"B_z_GSE": 999.9,
	"B_y_GSM": 999.9,
	"B_z_GSM": 999.9,
	"sigma_B_mag_avg": 999.9,
	"sigma_B_mag": 999.9,
	"sigma_B_x_GSE": 999.9,
	"sigma_B_y_GSE": 999.9,
	"sigma_B_z_GSE": 999.9,
	"T_p": 9999999.0,
	"n_p": 999.9,
	"v_plasma": 9999.0,
	"phi_v": 999.9,
	"theta_v": 999.9,
	"n_alpha_n_p": 9.999,
	"p_flow": 99.99,
	"sigma_T": 9999999.0,
	"sigma_n": 999.9,
	"sigma_v": 9999.0,
	"sigma_phi_v": 999.9,
	"sigma_theta_v": 999.9,
	"sigma_na_np": 9.999,
	"E": 999.99,
	"beta_plasma": 999.99,
	"mach": 999.9,
	"Kp": 9.9,
	"R": 999,
	"Dst": 99999,
	"AE": 9999,
	"p_01MeV": 999999.99,
	"p_02MeV": 99999.99,
	"p_04MeV": 99999.99,
	"p_10MeV": 99999.99,
	"p_30MeV": 99999.99,
	"p_60MeV": 99999.99,
	"flag": 0,
	"Ap": 999,
	"f107_adj": 999.9,
	"PC": 999.9,
	"AL": 99999,
	"AU": 99999,
	"mach_mag": 99.9,
	"Lya": 0.999999,
	"QI_p": 9.9999
}


def _doc_param(**sub):
	def dec(obj):
		obj.__doc__ = obj.__doc__.format(**sub)
		return obj
	return dec


@_doc_param(prefix=OMNI_PREFIX, ext=OMNI_EXT)
def cache_omnie(
	year,
	prefix=None,
	ext=None,
	local_path=None,
	url_base=None,
):
	"""Download OMNI2 data to local cache

	Downloads the OMNI2 (extended) data file from [#]_ to the local location.

	.. [#] https://spdf.gsfc.nasa.gov/pub/data/omni/low_res_omni/extended/

	Parameters
	----------
	year: int
		Year of the data.
	prefix: `None` or str, optional, default `None`
		File prefix for constructing the file name as <prefix>_year.<ext>.
		`None` defaults to '{prefix}'.
	ext: `None` or str, optional, default `None`
		File extension for constructing the file name as <prefix>_year.<ext>.
		`None` defaults to '{ext}'.
	local_path: `None` or str, optional, default `None`
		Path to the locally stored data yearly files, defaults to the
		data location within the package.
		`None` uses the package's default file location.
	url_base: `None` or str, optional, default `None`
		URL for the directory that contains the yearly files.
		`None` uses the default base url.

	Returns
	-------
	Nothing.
	"""
	prefix = prefix or OMNI_PREFIX
	ext = ext or OMNI_EXT
	local_path = local_path or LOCAL_PATH
	url_base = url_base or OMNI_URL_BASE

	basename = "{0}_{1:04d}.{2}".format(prefix, year, ext)

	if not os.path.exists(local_path):
		os.makedirs(local_path)

	omnie_file = os.path.join(local_path, basename)
	if not os.path.exists(omnie_file):
		url = urljoin(url_base, basename)
		logging.info("%s not found, downloading from %s.", omnie_file, url)
		_dl_file(omnie_file, url)


def omnie_mask_missing(df):
	"""Mask missing values with NaN

	Marks missing values in the OMNI2 data set by NaN.
	The missing value indicating numbers are taken from the file format description
	https://spdf.gsfc.nasa.gov/pub/data/omni/low_res_omni/extended/aareadme_extended

	Parameters
	----------
	df: pandas.DataFrame
		The OMNI2 data set, e.g. from :func:`omnie_hourly()` or :func:`read_omnie()`.

	Returns
	-------
	df: pandas.DataFrame
		The same dataframe with the missing values masked with ``numpy.nan``.

	Notes
	-----
	This function returns a copy of the dataframe, and all the integer columns
	will be converted to float to support NaN.
	"""
	res = df.copy()
	for _c in df.columns:
		_m = _OMNI_MISSING.get(_c, None)
		if _m is None:
			continue
		_mask = df[_c] != _m
		res[_c] = df[_c].where(_mask)
	return res


def read_omnie(omnie_file):
	"""Read and parse OMNI2 extended files [#]_

	Parses the Omni2 extended data files,  available at [#]_,
	into a :class:`pandas.DataFrame`.

	.. [#] https://omniweb.gsfc.nasa.gov/ow.html
	.. [#] https://spdf.gsfc.nasa.gov/pub/data/omni/low_res_omni/extended/

	Parameters
	----------
	omnie_file: str
		File to parse, absolute path or relative to the current dir.

	Returns
	-------
	sw_df: pandas.DataFrame
		The parsed OMNI2 space weather data (hourly values).
		Details in
		https://spdf.gsfc.nasa.gov/pub/data/omni/low_res_omni/extended/aareadme_extended

		Raises an ``IOError`` if the file is not found.
		The index is returned timezone-naive but contains UTC timestamps.
		To convert to a timezone-aware index, use
		:meth:`pandas.DataFrame.tz_localize()`: ``sw_df.tz_localize("utc")``.

		The dataframe contains the following columns:

		year:
			The observation year
		doy:
			Day of the year
		hour:
			Hour of the day
		bsrn:
			Bartels Solar Rotation Number.
		id_imf:
			ID for IMF spacecraft
		id_sw:
			ID for SW plasma spacecraft
		n_imf:
			Number of points in IMF averages
		n_plasma:
			Numberof points in plasma averages
		B_mag_avg:
			Magnetic field magnitude average B
		B_mag:
			Magnetic field vector magnitude
		theta_B:
			Latitude angle of the magnetic field vector
		phi_B:
			Longitude angle of the magnetic field vector
		B_x:
			B_x GSE, GSM
		B_y_GSE:
			B_y GSE
		B_z_GSE:
			B_z GSE
		B_y_GSM:
			B_y GSM
		B_z_GSM:
			B_z GSM
		sigma_B_mag_avg:
			RMS standard deviation of B_mag_avg
		sigma_B_mag:
			RMS standard deviation of B_mag
		sigma_B_x_GSE:
			RMS standard deviation of B_x_GSE
		sigma_B_y_GSE:
			RMS standard deviation of B_y_GSE
		sigma_B_z_GSE:
			RMS standard deviation of B_z_GSE
		T_p:
			Proton temperature
		n_p:
			Proton density
		v_plasma:
			Plasma flow speed
		phi_v:
			Plasma flow longitude angle
		theta_v:
			Plasma flow latitude angle
		n_alpha_n_p:
			Alpha/Proton ratio
		p_flow:
			Flow pressure
		sigma_T:
			Standard deviation of T_p
		sigma_n:
			Standard deviation of n_p
		sigma_v:
			Standard deviation of v_plasma
		sigma_phi_v:
			Standard deviation of phi_v
		sigma_theta_v:
			Standard deviation of theta_v
		sigma_na_np:
			Standard deviation of n_alpha_n_p
		E:
			Electric field magnitude
		beta_plasma:
			Plasma beta
		mach:
			Alfvén Mach number
		Kp:
			Kp index value
		R:
			Sunspot number
		Dst:
			Dst index value
		AE:
			AE index value
		p_01MeV, p_02MeV, p_04MeV, p_10MeV, p_30MeV, p_60MeV:
			Proton fluxes >1 MeV, >2 MeV, >4 MeV, >10 MeV, >30 MeV, > 60 MeV
		flag:
			Flag (-1, ..., 6)
		Ap:
			Ap index value
		f107_adj:
			F10.7 radio flux at 1 AU
		PC:
			PC index value
		AL, AU:
			AL and AU index values
		mach_mag:
			Magnetosonic Mach number

		The extended dataset contains the addional columns:

		Lya:
			Solar Lyman-alpha irradiance
		QI_p:
			Proton QI
	"""
	_assert_file_exists(omnie_file)
	# FORMAT(
	#     2I4,I3,I5,2I3,2I4,14F6.1,F9.0,F6.1,F6.0,2F6.1,F6.3,F6.2,
	#     F9.0,F6.1,F6.0,2F6.1,F6.3,2F7.2,F6.1,I3,I4,I6,I5,F10.2,
	#     5F9.2,I3,I4,2F6.1,2I6,F5.1,F9.6,F7.4
	# )
	sw = np.genfromtxt(
		omnie_file,
		skip_header=0,
		delimiter=[
		#   1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20
		#  yy dd hr br i1 i2 n1 n2  B B' tB fB Bx By Bz By Bz sB sB sB
			4, 4, 3, 5, 3, 3, 4, 4, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
		#  21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40
		#  sB sB Tp np  v fv tv nr  p sT sn sv sf st sr  E bp  M Kp  R
			6, 6, 9, 6, 6, 6, 6, 6, 6, 9, 6, 6, 6, 6, 6, 7, 7, 6, 3, 4,
		#  41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57
		#  Ds AE p1 p2 p4p10p30p60 fl Apf10 PC AL AU Mm La QI
			6, 5,10, 9, 9, 9, 9, 9, 3, 4, 6, 6, 6, 6, 5, 9, 7,
		],
		dtype=(
			"i4,i4,i4,i4,i4,i4,i4,i4,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,"
			"f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,f8,i4,i4,"
			"i4,i4,f8,f8,f8,f8,f8,f8,i4,i4,f8,f8,i4,i4,f8,f8,f8"
		),
		names=[
			"year", "doy", "hour", "bsrn", "id_imf", "id_sw", "n_imf", "n_plasma",
			"B_mag_avg", "B_mag", "theta_B", "phi_B",
			"B_x", "B_y_GSE", "B_z_GSE", "B_y_GSM", "B_z_GSM",
			"sigma_B_mag_avg", "sigma_B_mag",
			"sigma_B_x_GSE", "sigma_B_y_GSE", "sigma_B_z_GSE",
			"T_p", "n_p", "v_plasma", "phi_v", "theta_v", "n_alpha_n_p", "p_flow",
			"sigma_T", "sigma_n", "sigma_v",
			"sigma_phi_v", "sigma_theta_v", "sigma_na_np",
			"E", "beta_plasma", "mach", "Kp", "R", "Dst", "AE",
			"p_01MeV", "p_02MeV", "p_04MeV", "p_10MeV", "p_30MeV", "p_60MeV",
			"flag", "Ap", "f107_adj", "PC", "AL", "AU", "mach_mag", "Lya", "QI_p",
		]
	)
	sw = sw[sw["year"] != -1]
	ts = pd.to_datetime(
		[
			"{0:04d}.{1:03d} {2:02d}".format(yy, dd, hh)
			for yy, dd, hh in sw[["year", "doy", "hour"]]
		],
		format="%Y.%j %H",
	)
	sw_df = pd.DataFrame(sw, index=ts)
	# Adjust Kp to 0...9
	sw_df["Kp"] = 0.1 * sw_df["Kp"]
	return sw_df


@_doc_param(prefix=OMNI_PREFIX, ext=OMNI_EXT)
def omnie_hourly(
	year,
	prefix=None,
	ext=None,
	local_path=None,
	url_base=None,
	cache=False,
):
	"""OMNI hourly data for year `year`

	Loads the OMNI hourly data for the given year,
	from the locally cached data.
	Use `local_path` to set a custom location if you
	have the omni data already available.

	Parameters
	----------
	year: int
		Year of the data.
	prefix: `None` or str, optional, default `None`
		File prefix for constructing the file name as <prefix>_year.<ext>.
		`None` defaults to '{prefix}'.
	ext: `None` or str, optional, default `None`
		File extension for constructing the file name as <prefix>_year.<ext>.
		`None` defaults to '{ext}'.
	local_path: `None` or str, optional, default `None`
		Path to the locally stored data yearly files, defaults to the
		data location within the package.
		`None` uses the package's default file location.
	url_base: `None` or str, optional, default `None`
		URL for the directory that contains the yearly files.
		`None` uses the default base url.
	cache: boolean, optional, default False
		Download files locally if they are not already available.

	Returns
	-------
	sw_df: pandas.DataFrame
		The parsed space weather data (hourly values).

		Raises an ``IOError`` if the file is not available.
		The index is returned timezone-naive but contains UTC timestamps.
		To convert to a timezone-aware index, use
		:meth:`pandas.DataFrame.tz_localize()`: ``sw_df.tz_localize("utc")``.

	See Also
	--------
	read_omnie
	"""
	prefix = prefix or OMNI_PREFIX
	ext = ext or OMNI_EXT
	local_path = local_path or LOCAL_PATH
	url_base = url_base or OMNI_URL_BASE

	basename = "{0}_{1:04d}.{2}".format(prefix, year, ext)
	omnie_file = os.path.join(local_path, basename)

	# ensure that the file exists
	if not os.path.exists(omnie_file):
		warn("Could not find OMNI2 data {0}.".format(omnie_file))
		if cache:
			cache_omnie(
				year,
				prefix=prefix, ext=ext,
				local_path=local_path, url_base=url_base,
			)
		else:
			warn(
				"Local data files not found, pass `cache=True` "
				"or run `sw.cache_omnie()` to download the file."
			)

	return read_omnie(omnie_file)



from codecs import open
from os import path
import re
# Always prefer setuptools over distutils
from setuptools import find_packages, setup

name = "spaceweather"
meta_path = path.join("src", name, "__init__.py")
here = path.abspath(path.dirname(__file__))

extras_require = {
	"tests": ["pytest", "pytest-mock"],
}
extras_require["all"] = sorted(
	{v for req in extras_require.values() for v in req})


# Approach taken from
# https://packaging.python.org/guides/single-sourcing-package-version/
# and the `attrs` package https://www.attrs.org/
# https://github.com/python-attrs/attrs
def read(*parts):
	"""
	Builds an absolute path from *parts* and and return the contents of the
	resulting file.  Assumes UTF-8 encoding.
	"""
	with open(path.join(here, *parts), "rb", "utf-8") as f:
		return f.read()


def find_meta(meta, *path):
	"""
	Extract __*meta*__ from *path* (can have multiple components)
	"""
	meta_file = read(*path)
	meta_match = re.search(
		r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta), meta_file, re.M,
	)
	if not meta_match:
		raise RuntimeError("__{meta}__ string not found.".format(meta=meta))
	return meta_match.group(1)


# Get the long description from the README file
long_description = read("README.md")
version = find_meta("version", meta_path)

if __name__ == "__main__":
	setup(
		name=name,
		version=version,
		description="Python interface for space weather indices",
		long_description=long_description,
		long_description_content_type="text/markdown",
		url="https://github.com/st-bender/pyspaceweather",
		author="Stefan Bender",
		author_email="sbender@iaa.es",
		license="GPLv2",
		classifiers=[
			"Development Status :: 4 - Beta",
			"Intended Audience :: Science/Research",
			"License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
			"Programming Language :: Python",
			"Programming Language :: Python :: 2",
			"Programming Language :: Python :: 3",
			"Programming Language :: Python :: Implementation :: CPython",
			"Topic :: Scientific/Engineering :: Physics",
			"Topic :: Utilities",
		],
		packages=find_packages("src"),
		package_dir={"": "src"},
		package_data={
			"spaceweather": ["data/.cache", "data/omni_extended/.cache"],
		},
		include_package_data=True,
		install_requires=[
			"numpy>=1.13.0",
			"pandas",
			"requests",
		],
		extras_require=extras_require,
		scripts=[],
		entry_points={},
		zip_safe=False,
	)



# Copyright (c) 2020--2024 Stefan Bender
#
# This module is part of pyspaceweather.
# pyspaceweather is free software: you can redistribute it or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation, version 2.
# See accompanying COPYING.GPLv2 file or http://www.gnu.org/licenses/gpl-2.0.html.
"""Python interface for space weather indices from GFZ Potsdam

GFZ space weather indices ASCII file parser for python [#]_.
Includes parser for the GFZ ASCII files, files in WDC format,
and for the Hpo 30 and 60 minute ASCII files [#]_.
For the file formats, see the `gfz_xxx_format.txt` files.

.. [#] https://kp.gfz-potsdam.de/en/
.. [#] https://kp.gfz-potsdam.de/en/hp30-hp60
"""
import os
import logging
from warnings import warn

import numpy as np
import pandas as pd

from .core import _assert_file_exists, _dl_file, _resource_filepath

__all__ = [
	"gfz_daily", "gfz_3h", "read_gfz",
	"read_gfz_hp",
	"get_gfz_age", "update_gfz",
	"update_gfz_hp30", "update_gfz_hp60",
	"GFZ_PATH_ALL", "GFZ_PATH_30D",
	"HP30_PATH_ALL", "HP30_PATH_30D",
	"HP60_PATH_ALL", "HP60_PATH_30D",
]

GFZ_URL_ALL = "https://kp.gfz-potsdam.de/app/files/Kp_ap_Ap_SN_F107_since_1932.txt"
GFZ_URL_30D = "https://kp.gfz-potsdam.de/app/files/Kp_ap_Ap_SN_F107_nowcast.txt"
GFZ_FILE_ALL = os.path.basename(GFZ_URL_ALL)
GFZ_FILE_30D = os.path.basename(GFZ_URL_30D)
GFZ_PATH_ALL = _resource_filepath(GFZ_FILE_ALL)
GFZ_PATH_30D = _resource_filepath(GFZ_FILE_30D)

HP30_URL_ALL = "https://kp.gfz.de/app/files/Hp30_ap30_complete_series.txt"
HP30_URL_30D = "https://kp.gfz.de/app/files/Hp30_ap30_nowcast.txt"
HP30_FILE_ALL = os.path.basename(HP30_URL_ALL)
HP30_FILE_30D = os.path.basename(HP30_URL_30D)
HP30_PATH_ALL = _resource_filepath(HP30_FILE_ALL)
HP30_PATH_30D = _resource_filepath(HP30_FILE_30D)

HP60_URL_ALL = "https://kp.gfz.de/app/files/Hp60_ap60_complete_series.txt"
HP60_URL_30D = "https://kp.gfz.de/app/files/Hp60_ap60_nowcast.txt"
HP60_FILE_ALL = os.path.basename(HP60_URL_ALL)
HP60_FILE_30D = os.path.basename(HP60_URL_30D)
HP60_PATH_ALL = _resource_filepath(HP60_FILE_ALL)
HP60_PATH_30D = _resource_filepath(HP60_FILE_30D)


def get_gfz_age(gfzpath, relative=True):
	"""Age of the downloaded data file

	Retrieves the last update time of the given file or full path.

	Parameters
	----------
	gfzpath: str
		Filename to check, absolute path or relative to the current dir.
	relative: bool, optional, default True
		Return the file's age (True) or the last update time (False).

	Returns
	-------
	upd: pandas.Timestamp or pandas.Timedelta
		The last updated time or the file age, depending on the setting
		of `relative` above.
		Raises ``IOError`` if the file is not found.
	"""
	_assert_file_exists(gfzpath)
	with open(gfzpath) as fp:
		for line in fp:
			# forward to last line
			pass
	upd = pd.to_datetime(line[:10].replace(" ", "-"), utc=True)
	if relative:
		return pd.Timestamp.utcnow() - upd
	return upd


def update_gfz(
	min_age="1d",
	gfzpath_all=None, gfzpath_30d=None,
	url_all=None, url_30d=None,
):
	"""Update the local space weather index data

	Updates the local space weather index data from the website [#]_,
	given that the 30-day file is older than `min_age`,
	or the combined (large) file is older than 30 days.
	If the data is missing for some reason, a download will be attempted nonetheless.

	All arguments are optional and changing them from the defaults should
	neither be necessary nor is it recommended.

	.. [#] https://kp.gfz-potsdam.de/en/

	Parameters
	----------
	min_age: str, optional, default "1d"
		The time after which a new download will be attempted.
		The online data is updated every day, thus setting this value to
		a shorter time is not needed and not recommended.
	gfzpath_all: `None` or str, optional, default `None`
		Filename for the large combined index file including the
		historic data, absolute path or relative to the current dir.
		`None` uses the package's default file location.
	gfzpath_30d: `None` or str, optional, default `None`
		Filename for the 30-day (nowcast) index file, absolute path or relative
		to the current dir.
		`None` uses the package's default file location.
	url_all: `None` or str, optional, default `None`
		The url of the "historic" data file.
		`None` uses the default url.
	url_30d: `None` or str, optional, default `None`
		The url of the data file containing the indices for the last 30 days.
		`None` uses the default url.

	Returns
	-------
	Nothing.
	"""
	def _update_file(gfzpath, url, min_age):
		if not os.path.exists(gfzpath):
			logging.info("{0} not found, downloading.".format(gfzpath))
			_dl_file(gfzpath, url)
			return
		if get_gfz_age(gfzpath) < pd.Timedelta(min_age):
			logging.info("not updating '{0}'.".format(gfzpath))
			return
		logging.info("updating '{0}'.".format(gfzpath))
		_dl_file(gfzpath, url)

	gfzpath_all = gfzpath_all or GFZ_PATH_ALL
	gfzpath_30d = gfzpath_30d or GFZ_PATH_30D
	url_all = url_all or GFZ_URL_ALL
	url_30d = url_30d or GFZ_URL_30D

	# Update the large file after 30 days
	_update_file(gfzpath_all, url_all, "30days")
	# Don't re-download before `min_age` has passed (1d)
	_update_file(gfzpath_30d, url_30d, min_age)


def update_gfz_hp30(
	min_age="1d",
	gfzpath_all=None, gfzpath_30d=None,
	url_all=None, url_30d=None,
):
	"""Updates the local Hp30 index data

	See Also
	--------
	update_gfz
	"""
	gfzpath_all = gfzpath_all or HP30_PATH_ALL
	gfzpath_30d = gfzpath_30d or HP30_PATH_30D
	url_all = url_all or HP30_URL_ALL
	url_30d = url_30d or HP30_URL_30D
	return update_gfz(
		min_age=min_age,
		gfzpath_all=gfzpath_all, gfzpath_30d=gfzpath_30d,
		url_all=url_all, url_30d=url_30d,
	)


def update_gfz_hp60(
	min_age="1d",
	gfzpath_all=None, gfzpath_30d=None,
	url_all=None, url_30d=None,
):
	"""Updates the local Hp60 index data

	See Also
	--------
	update_gfz
	"""
	gfzpath_all = gfzpath_all or HP60_PATH_ALL
	gfzpath_30d = gfzpath_30d or HP60_PATH_30D
	url_all = url_all or HP60_URL_ALL
	url_30d = url_30d or HP60_URL_30D
	return update_gfz(
		min_age=min_age,
		gfzpath_all=gfzpath_all, gfzpath_30d=gfzpath_30d,
		url_all=url_all, url_30d=url_30d,
	)


def read_gfz(gfzpath):
	"""Read and parse space weather index data file

	Reads the given file and parses it according to the space weather data format.

	Parameters
	----------
	gfzpath: str
		File to parse, absolute path or relative to the current dir.

	Returns
	-------
	gfz_df: pandas.DataFrame
		The parsed space weather data (daily values).
		Raises an ``IOError`` if the file is not found.

		The dataframe contains the following columns:

		"year", "month", "day":
			The observation date
		"bsrn":
			Bartels Solar Rotation Number.
			A sequence of 27-day intervals counted continuously from 1832 Feb 8.
		"rotd":
			Number of Day within the Bartels 27-day cycle (01-27).
		"Kp0", "Kp3", "Kp6", "Kp9", "Kp12", "Kp15", "Kp18", "Kp21":
			Planetary 3-hour Range Index (Kp) for 0000-0300, 0300-0600,
			0600-0900, 0900-1200, 1200-1500, 1500-1800, 1800-2100, 2100-2400 UT
		"Kpsum": Sum of the 8 Kp indices for the day.
			Expressed to the nearest third of a unit.
		"Ap0", "Ap3", "Ap6", "Ap9", "Ap12", "Ap15", "Ap18", "Ap21":
			Planetary Equivalent Amplitude (Ap) for 0000-0300, 0300-0600,
			0600-0900, 0900-1200, 1200-1500, 1500-1800, 1800-2100, 2100-2400 UT
		"Apavg":
			Arithmetic average of the 8 Ap indices for the day.
		"isn":
			International Sunspot Number.
			Records contain the Zurich number through 1980 Dec 31 and the
			International Brussels number thereafter.
		"f107_obs":
			Observed (unadjusted) value of F10.7.
		"f107_adj":
			10.7-cm Solar Radio Flux (F10.7) Adjusted to 1 AU.
			Measured at Ottawa at 1700 UT daily from 1947 Feb 14 until
			1991 May 31 and measured at Penticton at 2000 UT from 1991 Jun 01 on.
			Expressed in units of 10-22 W/m2/Hz.
		"D":
			Definitive indicator.
			0: Kp and SN preliminary
			1: Kp definitive, SN preliminary
			2: Kp and SN definitive
	"""
	_assert_file_exists(gfzpath)
	gfz = np.genfromtxt(
		gfzpath,
		skip_header=3,
		delimiter=[
		#  yy mm dd dd dm br db kp kp kp kp kp kp kp kp
			4, 3, 3, 6, 8, 5, 3, 7, 7, 7, 7, 7, 7, 7, 7,
		#  ap ap ap ap ap ap ap ap Ap sn f1 f2 def
			5, 5, 5, 5, 5, 5, 5, 5, 6, 4, 9, 9, 2,
		],
		dtype=(
			"i4,i4,i4,i4,f4,i4,i4,f4,f4,f4,f4,f4,f4,f4,"
			"f4,i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,f8,f8,i4,"
		),
		names=[
			"year", "month", "day", "days", "days_m", "bsrn", "rotd",
			"Kp0", "Kp3", "Kp6", "Kp9", "Kp12", "Kp15", "Kp18", "Kp21",
			"Ap0", "Ap3", "Ap6", "Ap9", "Ap12", "Ap15", "Ap18", "Ap21", "Apavg",
			"isn", "f107_obs", "f107_adj", "D",
		]
	)
	gfz = gfz[gfz["year"] != -1]
	ts = pd.to_datetime([
		"{0:04d}-{1:02d}-{2:02d}".format(yy, mm, dd)
		for yy, mm, dd in gfz[["year", "month", "day"]]
	])
	gfz_df = pd.DataFrame(gfz, index=ts)
	# Sum Kp for compatibility with celestrak dataframe
	kpns = list(map("Kp{0}".format, range(0, 23, 3)))
	gfz_df.insert(15, "Kpsum", gfz_df[kpns].sum(axis=1))
	return gfz_df


def read_gfz_hp(gfzhppath):
	"""Read and parse GFZ Hp30 and Hp60 index data file

	Reads the given file and parses it according to the Hp30 and Hp60 file format.
	File format descriptions in [#]_ and [#]_

	.. [#] https://kp.gfz-potsdam.de/app/format/Hpo_Hp30.txt
	.. [#] https://kp.gfz-potsdam.de/app/format/Hpo_Hp60.txt

	Parameters
	----------
	gfzhppath: str
		File to parse, absolute path or relative to the current dir.

	Returns
	-------
	hp_df: pandas.DataFrame
		The parsed space weather data with the 30 min or 60 min index.
		The index is returned timezone-naive but contains UTC timestamps.
		To convert to a timezone-aware index, use
		:meth:`pandas.DataFrame.tz_localize()`: ``hp_df.tz_localize("utc")``.
		Raises an ``IOError`` if the file is not found.

		The dataframe contains the following columns:

		"index":
			padas.DateTimeIndex of the middle times of the intervals.
		"year", "month", "day":
			The observation date.
		"hh_h":
			Starting time in hours of interval.
		"hh_m":
			Middle time in hours of interval.
		"days":
			Days since 1932-01-01 00:00 UT to start of interval.
		"days_m":
			Days since 1932-01-01 00:00 UT to middle of interval.
		"Hp":
			Hp index during to the interval (30 min or 60 min).
		"ap":
			ap index during to the interval (30 min or 60 min).
		"D":
			Reserved for future use, D = 0 for now.
	"""
	_assert_file_exists(gfzhppath)
	hp = np.genfromtxt(
		gfzhppath,
		delimiter=[
		#  yy mm dd hh hm ddd ddm hp ap  D
			4, 3, 3, 5, 6, 12, 12, 7, 5, 2,
		],
		dtype=(
			"i4,i4,i4,f4,f4,f4,f4,f4,i4,i4"
		),
		names=[
			"year", "month", "day", "hh_h", "hh_m", "days", "days_m", "Hp", "ap", "D",
		]
	)
	hp = hp[hp["year"] != -1]
	ts = pd.to_datetime([
		"{0:04d}-{1:02d}-{2:02d} {3:02d}:{4:02d}".format(
			yy, mm, dd, int(np.floor(hh_m)), int(60 * (hh_m - np.floor(hh_m)))
		)
		for yy, mm, dd, hh_m in hp[["year", "month", "day", "hh_m"]]
	])
	hp_df = pd.DataFrame(hp, index=ts)
	return hp_df


def read_gfz_wdc(gfzpath):
	"""Parse space weather index data file in WDC format

	Parses the GFZ index data in WDC format.

	Parameters
	----------
	gfzpath: str
		File to parse, absolute path or relative to the current dir.

	Returns
	-------
	gfz_df: pandas.DataFrame
		The parsed space weather data (daily values).
		Raises an ``IOError`` if the file is not found.
		The index is returned timezone-naive but contains UTC timestamps.
		To convert to a timezone-aware index, use
		:meth:`pandas.DataFrame.tz_localize()`: ``gfz_df.tz_localize("utc")``.

		The dataframe contains the following columns:

		"year", "month", "day":
			The observation date
		"bsrn":
			Bartels Solar Rotation Number.
			A sequence of 27-day intervals counted continuously from 1832 Feb 8.
		"rotd":
			Number of Day within the Bartels 27-day cycle (01-27).
		"Kp0", "Kp3", "Kp6", "Kp9", "Kp12", "Kp15", "Kp18", "Kp21":
			Planetary 3-hour Range Index (Kp) for 0000-0300, 0300-0600,
			0600-0900, 0900-1200, 1200-1500, 1500-1800, 1800-2100, 2100-2400 UT
		"Kpsum": Sum of the 8 Kp indices for the day.
			Expressed to the nearest third of a unit.
		"Ap0", "Ap3", "Ap6", "Ap9", "Ap12", "Ap15", "Ap18", "Ap21":
			Planetary Equivalent Amplitude (Ap) for 0000-0300, 0300-0600,
			0600-0900, 0900-1200, 1200-1500, 1500-1800, 1800-2100, 2100-2400 UT
		"Apavg":
			Arithmetic average of the 8 Ap indices for the day.
		"Cp":
			Cp index - the daily planetary character figure, a qualitative
			estimate of the overall level of geomagnetic activity for this day
			determined from the sum of the eight ap amplitudes,
			ranging from 0.0 to 2.5 in steps of 0.1.
		"C9":
			The contracted scale for Cp with only 1 digit, from 0 to 9.
	"""
	_assert_file_exists(gfzpath)
	gfz = np.genfromtxt(
		gfzpath,
		skip_header=3,
		delimiter=[
		#  yy mm dd br db kp kp kp kp kp kp kp kp kps
			2, 2, 2, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3,
		#  ap ap ap ap ap ap ap ap Ap Cp C9
			3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 1,
		],
		dtype=(
			"i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,i4,"
			"i4,i4,i4,i4,i4,i4,i4,i4,i4,f8,i4,"
		),
		names=[
			"year", "month", "day", "bsrn", "rotd",
			"Kp0", "Kp3", "Kp6", "Kp9", "Kp12", "Kp15", "Kp18", "Kp21", "Kpsum",
			"Ap0", "Ap3", "Ap6", "Ap9", "Ap12", "Ap15", "Ap18", "Ap21", "Apavg",
			"Cp", "C9",
		]
	)
	gfz = gfz[gfz["year"] != -1]
	ts = pd.to_datetime([
		"{0:04d}-{1:02d}-{2:02d}".format(2000 + yy if yy < 32 else 1900 + yy, mm, dd)
		for yy, mm, dd in gfz[["year", "month", "day"]]
	])
	gfz_df = pd.DataFrame(gfz, index=ts)
	gfz_df.loc[:, "year"] = ts.year
	# Adjust Kp to 0...9
	kpns = list(map("Kp{0}".format, range(0, 23, 3))) + ["Kpsum"]
	gfz_df[kpns] = 0.1 * gfz_df[kpns]
	return gfz_df


# Common arguments for the public daily and 3h interfaces
_GFZ_COMMON_PARAMS = """
Parameters
----------
gfzpath_all: `None` or str, optional, default `None`
	Filename for the large combined index file including the
	historic data, absolute path or relative to the current dir.
	`None` uses the package's default file location.
gfzpath_30d: `None` or str, optional, default `None`
	Filename for the 30-day (nowcast) index file,
	absolute path or relative to the current dir.
	`None` uses the package's default file location.
update: bool, optional, default False
	Attempt to update the local data if it is older than `update_interval`.
update_interval: str, optional, default "10days"
	The time after which the data are considered "old".
	By default, no automatic re-download is initiated, set `update` to true.
	The online data is updated every 3 hours, thus setting this value to
	a shorter time is not needed and not recommended.
gfz_format: `None` or str, optional, default `None`
	The file format to parse the files passed via `gfzpath_all` and `gfzpath_all`.
	Use `None`, "default", "gfz", or "standard" for the "standard" GFZ ASCII files.
	Use "wdc" to parse files in WDC format into a full-length `pandas.DataFrame`.
	Use "hp30" or "hp60" to read the Hp30 and Hp60 data files.
"""

_PARSERS = {
	"default": (read_gfz, update_gfz),
	"gfz": (read_gfz, update_gfz),
	"standard": (read_gfz, update_gfz),
	"wdc": (read_gfz_wdc, update_gfz),
	"hp30": (read_gfz_hp, update_gfz_hp30),
	"hp60": (read_gfz_hp, update_gfz_hp60),
}


def _doc_param(**sub):
	def dec(obj):
		obj.__doc__ = obj.__doc__.format(**sub)
		return obj
	return dec


@_doc_param(params=_GFZ_COMMON_PARAMS)
def gfz_daily(
	gfzpath_all=None,
	gfzpath_30d=None,
	update=False,
	update_interval="10days",
	gfz_format=None,
):
	"""Combined daily Ap, Kp, and f10.7 index values

	Combines the "historic" and last-30-day data into one dataframe.

	All arguments are optional and changing them from the defaults should not
	be required neither should it be necessary nor is it recommended.
	{params}
	Returns
	-------
	gfz_df: pandas.DataFrame
		The combined parsed space weather data (daily values).
		Raises ``IOError`` if the data files cannot be found.
		The index is returned timezone-naive but contains UTC timestamps.
		To convert to a timezone-aware index, use
		:meth:`pandas.DataFrame.tz_localize()`: ``gfz_df.tz_localize("utc")``.

	See Also
	--------
	gfz_3h, read_gfz
	"""
	gfzpath_all = gfzpath_all or GFZ_PATH_ALL
	gfzpath_30d = gfzpath_30d or GFZ_PATH_30D
	gfz_format = gfz_format or "gfz"
	parse_func, update_func = _PARSERS[gfz_format.lower()]
	# ensure that the file exists and is up to date
	if (
		not os.path.exists(gfzpath_all)
		or not os.path.exists(gfzpath_30d)
	):
		warn("Could not find space weather data, trying to download.")
		update_func(gfzpath_all=gfzpath_all, gfzpath_30d=gfzpath_30d)

	if (
		get_gfz_age(gfzpath_all) > pd.Timedelta("30days")
		or get_gfz_age(gfzpath_30d) > pd.Timedelta(update_interval)
	):
		if update:
			update_func(gfzpath_all=gfzpath_all, gfzpath_30d=gfzpath_30d)
		else:
			warn(
				"Local data files are older than {0}, pass `update=True` or "
				"run `gfz.update_gfz()` manually if you need newer data.".format(
					update_interval
				)
			)

	df_all = parse_func(gfzpath_all)
	df_30d = parse_func(gfzpath_30d)
	return pd.concat([df_all, df_30d[df_all.index[-1]:].iloc[1:]])


@_doc_param(params=_GFZ_COMMON_PARAMS)
def gfz_3h(*args, **kwargs):
	"""3h values of Ap and Kp

	Provides the 3-hourly Ap and Kp indices from the full daily data set.

	Accepts the same arguments as `gfz_daily()`.
	All arguments are optional and changing them from the defaults should not
	be required neither should it be necessary nor is it recommended.
	{params}
	Returns
	-------
	gfz_df: pandas.DataFrame
		The combined Ap and Kp index data (3h values).
		The index values are centred at the 3h interval, i.e. at 01:30:00,
		04:30:00, 07:30:00, ... and so on.
		Raises ``IOError`` if the data files cannot be found.
		The index is returned timezone-naive but contains UTC timestamps.
		To convert to a timezone-aware index, use
		:meth:`pandas.DataFrame.tz_localize()`: ``gfz_df.tz_localize("utc")``.

	See Also
	--------
	gfz_daily
	"""
	daily_df = gfz_daily(*args, **kwargs)
	ret = daily_df.copy()
	apns = list(map("Ap{0}".format, range(0, 23, 3)))
	kpns = list(map("Kp{0}".format, range(0, 23, 3)))
	for i, (ap, kp) in enumerate(zip(apns, kpns)):
		ret[ap].index = daily_df[ap].index + pd.Timedelta((i * 3 + 1.5), unit="h")
		ret[kp].index = daily_df[kp].index + pd.Timedelta((i * 3 + 1.5), unit="h")
	gfz_ap = pd.concat(map(ret.__getitem__, apns))
	gfz_kp = pd.concat(map(ret.__getitem__, kpns))
	df = pd.DataFrame({"Ap": gfz_ap, "Kp": gfz_kp})
	return df.reindex(df.index.sort_values())
